︠47ad5e50-71b7-4fcd-8c36-3d8c674528f1︠
"""
Author: Stefan Keil, 2010-02-21.

This is an extended Version of
http://sage.math.washington.edu/sato-tate/
by William Stein.

Let E/K be an elliptic curve over a number field K, which is not the rational numbers QQ. For each finite place v of K, for which E has good reduction, one can calculate the number
a_v := q_v+1-|E(FF_v)|,
where q_v=Nv is the usual norm of v and |E(FF_v)| is the number of points of the mod v reduced curve over the finite residuo field FF_v := O_K / v, which is the ring of integers of K modulo v.
Hasse says |a_v| <= 2 \sqrt q_v,
so one can normalize a_v/(2 \sqrt q_v) = cos theta_v,
with an unique angle theta_v in [0,pi].
The Sato-Tate-Conjecture says, that if one runs over all finite places v of good reduction, then the theta_v (ordered by increasing norm Nv) are equidistributed with respect to the so called Sato-Tate-Measure 2/pi sin^2 theta d theta.

This Worksheet calculates the distribution of the angles theta_v of a given elliptic curve defined over a number field K up to all finite places v of good reduction of E, with norm Nv=q_v <= N, whereas the intervall [0,pi] will be divided into B bins. Additionally it draws the curve of the Sato-Tate-Measure for comparison.
"""
︡75d98f42-f37b-447f-8033-86facf374f28︡︡
︠10df15ff-7bfb-4cf4-a268-90d2ff64a9d8︠
# returns all theta_v of a given elliptic curve E over a number field K up to Nv=q_v<N
def sato_tate(E, K, N):
    all_theta_v = []
    for v in primes_of_K_up_to_norm_N(K,N):
        if E.has_good_reduction(v):
            all_theta_v.append(theta_of(E, v))
    return all_theta_v
︡12688307-01ab-4460-8d41-e72ebfd9e8c3︡︡
︠f18ce92d-a0ab-463c-acfc-df21c1f1c850︠
# calculates theta_v of a given elliptic curve E and prime v
def theta_of(E, v):
    E_red=E.reduction(v)
    q_v=v.norm()
    cardinality_of_E_red=E_red.cardinality()
    a_v=q_v+1-cardinality_of_E_red
    theta_v = acos(a_v/(2*sqrt(q_v)))
    return theta_v
︡ab503a5c-4946-4902-9a80-283b65f77d3e︡︡
︠47d379d4-68fd-46b7-984c-a752417d293f︠
# returns all primes v of a given number field K with Nv=q_v<N
def primes_of_K_up_to_norm_N(K, N):
    primes =[]
    for p in prime_range(N+1):
        principle_ideal_p = K.ideal(p)
        primfactors_of_principle_ideal_p = principle_ideal_p.prime_factors()
        # take only primes with norm < N+1
        for v in primfactors_of_principle_ideal_p:
            if v.norm()<N+1: 
                primes.append(v)
    return primes
︡f30af05a-a748-4537-8888-3cf9847a9515︡︡
︠6dfedead-d4a5-4cc4-ba34-30181d122511︠
# this function divides the intervall [0,pi] into B bins and then counts how many of the theta_v
# are in which bin, and then returns this distribution and the number of all theta_v
def distribution_of_all_theta_v_in_B_bins(all_theta_v, B):
    normalize = float(B/pi)
    vals = {}
    distribution = dict([(i,0) for i in range(B)])
    for theta_v in all_theta_v:
        # n is the number of the bin
        n = int(normalize*float(theta_v))
        # in case of theta_v=pi
        if n > B-1:
            n=B-1
        distribution[n] += 1
    return distribution, len(all_theta_v)
︡07aee728-d63f-4ce3-badc-3586bc332914︡︡
︠25fdf4ca-1ea7-4606-b584-7a8815249110︠
# this will make the blue recangles according to the given distribution, the number of bins B,
# and the total number of thetas
def graph(distribution, B, number_of_thetas):
    s = Graphics()
    left = float(0); right = float(pi)
    length = right - left
    w = length/B
    k = 0
    for i, n in distribution.iteritems():
        k += n
        # ith bin has n objects in it. 
        s += polygon([(w*i+left,0), (w*(i+1)+left,0), \
            (w*(i+1)+left, n/(number_of_thetas*w)), (w*i+left, n/(number_of_thetas*w))], \
            rgbcolor=(0,0,0.5))
    return s
︡3cffe445-9627-46da-8eca-20ff2590cfb6︡︡
︠b76b227e-a4a6-4045-94e9-2d0b0bc5e894︠
# plots curve of Sato-Tate measure
def sato_tate_measure():
    PI = float(pi)
    return plot(lambda x: (2/PI)*math.sin(x)^2, 0,pi, plot_points=200, \
        rgbcolor=(0.3,0.1,0.1), thickness=2)
︡44885c0c-32b9-4f81-998e-46b74ac9e042︡︡
︠8f183738-7104-4656-a84c-b08de266119a︠
# returns the plotted distribution of all theta_v with Nv=q_v<N of an elliptic curve E over K
# divided in B bins and adds the curve of the Sato-Tate measure for comparison
def graph_ellcurve(E, K, B, N):
    all_theta_v = sato_tate(E, K, N)
    distribution, number_of_thetas = distribution_of_all_theta_v_in_B_bins(all_theta_v,B)
    return graph(distribution, B, number_of_thetas) + sato_tate_measure()
︡79660050-9db7-4842-bbfe-13700691e203︡︡
︠39a51d9c-9792-4faa-a951-b7fe46e3ce84︠
# defines a number field K and an elliptic curve E over K
# CAUTION:
# K cannot be QQ
K.<i> = NumberField(x^2 + 1)
E = EllipticCurve(K,[2*i,3])
E
︡47ede665-0fbd-4e09-84e3-975feabab460︡{"stdout": "Elliptic Curve defined by y^2 = x^3 + 2*i*x + 3 over Number Field in i with defining polynomial x^2 + 1"}︡
︠85534310-d964-4d98-a9d1-4d05b2e1e6f6︠
# starts the program and shows the distribution of all theta_v with Nv=q_v<N 
# of an elliptic curve E over a number field K divided in B bins
g = graph_ellcurve(E,K,B=40,N=10000)
show(g)
︡7f49ba15-45f0-4fc3-823a-566fbd891f7e︡{"html": "<font color='black'><img src='cell://sage0.png'></font>"}︡
︠0efc9ade-2ef4-4521-93fe-20f324387ab8︠
# starts the program and shows the distribution of all theta_v with Nv=q_v<N 
# of an elliptic curve E over a number field K divided in B bins
g = graph_ellcurve(E,K,B=50,N=10000)
show(g)
︡4a4c24a4-a286-4ed1-92cd-8a7742bac3b5︡{"html": "<font color='black'><img src='cell://sage0.png'></font>"}︡
︠30684d68-df4c-4c38-99b7-e37deeb52ff7︠
K.<t> = NumberField(x^2 + 3)
E = EllipticCurve(K,[-5,-3*t])
E
︡b7397b0e-c805-450b-bac1-49d4e92eff2d︡{"stdout": "Elliptic Curve defined by y^2 = x^3 + (-5)*x + (-3*t) over Number Field in t with defining polynomial x^2 + 3"}︡
︠8f35d529-91de-4f9f-b722-82cee71e3390︠
E.global_integral_model()
︡59d11eea-5416-439f-82fb-d65b87c8a519︡{"stdout": "Elliptic Curve defined by y^2 = x^3 + (-5)*x + (-3*t) over Number Field in t with defining polynomial x^2 + 3"}︡
︠f5f4849f-0b79-41f3-989d-e5f67b6d6c7a︠
# starts the program and shows the distribution of all theta_v with Nv=q_v<N 
# of an elliptic curve E over a number field K divided in B bins
g = graph_ellcurve(E,K,B=50,N=10000)
show(g)
︡f2c9144d-9ec7-4c53-a3a3-c1cce31f878b︡{"html": "<font color='black'><img src='cell://sage0.png'></font>"}︡
︠48ce04ef-54e8-4ae4-82ca-cd371eab5c24︠
K.<t> = NumberField(x^2 + 167)
E = EllipticCurve(K,[t,2])
E
︡ef8f3f0c-3e6e-4c22-9ca2-683ddb5dbe6d︡{"stdout": "Elliptic Curve defined by y^2 = x^3 + t*x + 2 over Number Field in t with defining polynomial x^2 + 167"}︡
︠78674373-dc36-44ba-b442-be4a809b1b53︠
E.global_integral_model()
︡3ab7989f-14e2-41e2-a1cb-43116f7dc07c︡{"stdout": "Elliptic Curve defined by y^2 = x^3 + t*x + 2 over Number Field in t with defining polynomial x^2 + 167"}︡
︠11717cc9-64f5-48fe-bf3b-071ac19a8534︠
# starts the program and shows the distribution of all theta_v with Nv=q_v<N 
# of an elliptic curve E over a number field K divided in B bins
g = graph_ellcurve(E,K,B=50,N=10000)
show(g)
︡4bd75c98-17bc-4c73-b7ff-9bee27611136︡{"html": "<font color='black'><img src='cell://sage0.png'></font>"}︡
︠7b72ba73-e02f-4d2f-92c1-feeae6e017fc︠
K.<t> = NumberField(x^5 + 4*x^3 -17)
E = EllipticCurve(K,[t,1])
E
︡1e6e094c-721f-41e0-8aa7-902b0541fba1︡{"stdout": "Elliptic Curve defined by y^2 = x^3 + t*x + 1 over Number Field in t with defining polynomial x^5 + 4*x^3 - 17"}︡
︠cf19f21b-c9c0-4058-933d-db637e48fc47︠
E.global_integral_model()
︡ac26eb76-9012-4766-bf04-5f0b8df35c8c︡{"stdout": "Elliptic Curve defined by y^2 = x^3 + t*x + 1 over Number Field in t with defining polynomial x^5 + 4*x^3 - 17"}︡
︠9bce425d-2b3d-4cd8-9fd3-4813cbd78d5c︠
# starts the program and shows the distribution of all theta_v with Nv=q_v<N 
# of an elliptic curve E over a number field K divided in B bins
g = graph_ellcurve(E,K,B=50,N=10000)
show(g)
︡069ee12b-8177-45cc-b2c6-60399c17d04f︡{"html": "<font color='black'><img src='cell://sage0.png'></font>"}︡
